"""Conversion functions for Earth/D'Ni datetimes & Ordinal/Julian dates."""

from datetime import datetime, timezone
from math import floor
from typing import Dict, Optional, Tuple

from .constants import (
    DniDateTimeTuple,
    DNI_EPOCH_EDT,
    DNI_EPOCH_HAHR,
    DNI_UNITS,
    MAX_DNI_ORDINAL,
    MIN_DNI_ORDINAL,
    MS_PER_HAHR,
    MS_PER_PRORAHN,
    YAHRTEE_BEFORE_VAILEE
)

from .utils import (
    check_dni_date_fields,
    get_adj_ms_from_epoch,
    get_ms_from_epoch,
    is_timezone_aware
)


# Number of yahrtee (days) in a hahr (year) == 290.
_YAHR_IN_HAHR = DNI_UNITS["yahr"]["max"] * DNI_UNITS["vailee"]["max"]


def hvy2ord(hahr: int, vailee: int, yahr: int) -> int:
    """The amount of yahrtee since Leefo 1, 7654 (DniDate.min)."""
    ok_hahr, ok_vailee, ok_yahr = check_dni_date_fields(hahr, vailee, yahr)
    yahr_before_hahr = (ok_hahr - DNI_UNITS["hahr"]["min"]) * _YAHR_IN_HAHR
    return yahr_before_hahr + YAHRTEE_BEFORE_VAILEE[ok_vailee] + ok_yahr


def ord2hvy(ordinal: int) -> Tuple[int, int, int]:
    """Find the exact hahr, vailee, and yahr from an ordinal value of
    yahrtee.
    """
    if not MIN_DNI_ORDINAL <= ordinal <= MAX_DNI_ORDINAL:
        raise ValueError(
            f"Ordinal must be in {MIN_DNI_ORDINAL}..{MAX_DNI_ORDINAL}",
            ordinal,
        )
    # TODO: I hate this with a firey passion, but it works. Needs a revamp at
    # some point.
    hahr, rem = divmod(ordinal, _YAHR_IN_HAHR)
    hahr += DNI_UNITS["hahr"]["min"]
    if rem == 0:
        hahr -= 1
        vailee = DNI_UNITS["vailee"]["max"]
        yahr = DNI_UNITS["yahr"]["max"]
    else:
        vailee, yahr = divmod(
            rem + DNI_UNITS["yahr"]["max"],
            DNI_UNITS["yahr"]["max"]
        )
        if yahr == 0:
            vailee -= 1
            yahr = DNI_UNITS["yahr"]["max"]
    return hahr, vailee, yahr


def earth_to_dni(
    date_time: Optional[datetime] = None
) -> DniDateTimeTuple:
    """Converts an Earth gregorian datetime to a D'Ni datetime."""
    if date_time is None:
        moment = datetime.now(timezone.utc)
    else:
        if is_timezone_aware(date_time):
            moment = date_time.astimezone(timezone.utc)
        else:
            raise ValueError("Supplied datetime must be timezone aware.")

    dni_dt: Dict[str, int] = {}

    stamp = get_adj_ms_from_epoch(moment)
    dni_epoch = get_adj_ms_from_epoch(DNI_EPOCH_EDT)
    delta = float(stamp - dni_epoch)

    for name, details in DNI_UNITS.items():
        if name == "hahr":
            dni_dt[name] = floor(delta / MS_PER_HAHR)
            delta -= dni_dt[name] * MS_PER_HAHR
            delta *= details["total_pro"] / MS_PER_HAHR
        else:
            dni_dt[name] = floor(delta / details["total_pro"])
            delta -= dni_dt[name] * details["total_pro"]

    # Correct potential value underflow (dates before the reference date)
    temp_units = list(DNI_UNITS.items())
    rev_index = len(temp_units) - 1
    for index, unit in enumerate(reversed(temp_units)):
        if dni_dt[unit[0]] < 0:
            dni_dt[unit[0]] += unit[1]["max"]
            dni_dt[temp_units[rev_index - index][0]] -= 1

    # Add reference D'ni hahr
    dni_dt["hahr"] += DNI_EPOCH_HAHR

    # Vailee and Yahr are not 0-based
    dni_dt["vailee"] += 1
    dni_dt["yahr"] += 1

    return (
        dni_dt["hahr"],
        dni_dt["vailee"],
        dni_dt["yahr"],
        dni_dt["gahrtahvo"],
        dni_dt["tahvo"],
        dni_dt["gorahn"],
        dni_dt["prorahn"],
    )


def dni_to_earth(dni_date_time: DniDateTimeTuple) -> datetime:
    """Converts a D'Ni datetime to an Earth gregorian datetime."""

    if not isinstance(dni_date_time, tuple):
        raise ValueError(
            (
                "D'Ni datetime must be in the form of a timetuple. "
                "See DniDateTime.timetuple()."
            ),
            dni_date_time
        )

    if len(dni_date_time) != 7:
        raise ValueError(
            "The D'Ni timetuple is malformed. See DniDateTime.timetuple().",
            dni_date_time
        )

    temp_units = list(dni_date_time)

    temp_units[0] -= DNI_EPOCH_HAHR
    temp_units[1] -= 1
    temp_units[2] -= 1

    total_prorahn = 0

    for index, unit in enumerate(DNI_UNITS.items()):
        total_prorahn += temp_units[index] * unit[1]["total_pro"]

    timestamp = total_prorahn * MS_PER_PRORAHN
    epoch = get_ms_from_epoch(DNI_EPOCH_EDT)
    date = timestamp + epoch

    return datetime.fromtimestamp(date / 1000.0, timezone.utc)
