"""The DniDate class."""

from typing import Any, Optional, TYPE_CHECKING, Union

from .constants import DniDateTimeTuple, DNI_UNITS, MAX_DNI_ORDINAL
from .conversions import hvy2ord, ord2hvy
from .dnitimedelta import DniTimedelta
from .utils import check_dni_date_fields, cmp


class DniDate:
    """A D'Ni date."""

    __slots__ = "_hahr", "_vailee", "_yahr"

    if TYPE_CHECKING:
        _hahr: int
        _vailee: int
        _yahr: int

        # Class properties
        min: "DniDate"
        max: "DniDate"
        resolution: DniTimedelta

    def __new__(
        cls,
        hahr: int = DNI_UNITS["hahr"]["min"],
        vailee: int = DNI_UNITS["vailee"]["min"],
        yahr: int = DNI_UNITS["yahr"]["min"],
    ) -> "DniDate":
        if isinstance(hahr, int):
            hahr, vailee, yahr = check_dni_date_fields(hahr, vailee, yahr)
            self = object.__new__(cls)
            self._hahr = hahr
            self._vailee = vailee
            self._yahr = yahr
            return self
        return NotImplemented

    @classmethod
    def from_dnitimestamp(cls, prorahntee: int) -> "DniDate":
        """Construct a D'Ni date from a D'Ni timestamp."""
        yahrtee = prorahntee // DNI_UNITS["yahr"]["total_pro"]
        return cls.from_ordinal(yahrtee)

    @classmethod
    def today(cls) -> "DniDate":
        # TODO
        return NotImplemented

    @classmethod
    def from_isoformat(cls, date_str: str) -> "DniDate":
        if not isinstance(date_str, str):
            raise TypeError("from_isoformat: argument must be str.")

        try:
            assert 8 <= len(date_str) <= 11
            parts = date_str.split("-")
            if len(parts) != 3:
                raise ValueError(f"Invalid isoformat string: {date_str!r}")
            return cls(int(parts[0]), int(parts[1]), int(parts[2]))
        except Exception as exc:
            raise ValueError(f"Invalid isoformat string: {date_str!r}") from exc

    @classmethod
    def from_ordinal(cls, yahrtee: int) -> "DniDate":
        """Construct a D'ni date from a D'ni ordinal."""
        hahr, vailee, yahr = ord2hvy(yahrtee)
        return cls(hahr, vailee, yahr)

    def to_ordinal(self) -> int:
        """Return a D'ni ordinal for the hahr, vailee, and yahr."""
        return hvy2ord(self._hahr, self._vailee, self._yahr)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}"
            f"({self._hahr}, "
            f"{self._vailee}, "
            f"{self._yahr})"
        )

    def isoformat(self) -> str:
        """Return the date formatted according to ISO standard."""
        return f"{self._hahr}-{self._vailee:02}-{self._yahr:02}"

    __str__ = isoformat

    # Read-only field accessors
    @property
    def hahr(self) -> int:
        """hahr"""
        return self._hahr

    @property
    def vailee(self) -> int:
        """vailee"""
        return self._vailee

    @property
    def yahr(self) -> int:
        """yahr"""
        return self._yahr

    def replace(
        self,
        hahr: Optional[int] = None,
        vailee: Optional[int] = None,
        yahr: Optional[int] = None,
    ) -> "DniDate":
        """Return a new D'ni date with new values for the specified fields."""
        if hahr is None:
            hahr = self._hahr
        if vailee is None:
            vailee = self._vailee
        if yahr is None:
            yahr = self._yahr
        return type(self)(hahr, vailee, yahr)

    # Arithmetic
    def __add__(self, other: Any) -> "DniDate":
        if isinstance(other, DniTimedelta):
            ordinal = self.to_ordinal() + other._yahrtee
            if 0 < ordinal <= MAX_DNI_ORDINAL:
                return type(self).from_ordinal(ordinal)
            raise OverflowError("Result out of range.")
        return NotImplemented

    __radd__ = __add__

    def __sub__(self, other: Any) -> Union["DniDate", DniTimedelta]:
        if isinstance(other, DniTimedelta):
            return self + DniTimedelta(-other._yahrtee)
        if isinstance(other, DniDate):
            yahrtee1 = self.to_ordinal()
            yahrtee2 = other.to_ordinal()
            return DniTimedelta(yahrtee1 - yahrtee2)
        return NotImplemented

    # Rich comparisons
    def _cmp(self, other: Any) -> int:
        """Helper function for rich comparison of DniDate objects."""
        # pylint: disable=protected-access
        assert isinstance(other, DniDate)
        hahr1, vailee1, yahr1 = self._hahr, self._vailee, self._yahr
        hahr2, vailee2, yahr2 = other._hahr, other._vailee, other._yahr
        return cmp((hahr1, vailee1, yahr1), (hahr2, vailee2, yahr2))

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DniDate):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other: Any) -> bool:
        if isinstance(other, DniDate):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, DniDate):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, DniDate):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, DniDate):
            return self._cmp(other) > 0
        return NotImplemented

    def dnitimetuple(self) -> DniDateTimeTuple:
        """Local time tuple compatible with dnitime.localtime()."""
        # TODO
        return NotImplemented


DniDate.min = DniDate(7654, 1, 1)
DniDate.max = DniDate(17655, 8, 1)
DniDate.resolution = DniTimedelta(yahrtee=1)
