"""The DniTime class."""

from typing import Any, Optional, TYPE_CHECKING

from .constants import DNI_UNITS
from .dnitimedelta import DniTimedelta
from .utils import check_dni_time_fields, cmp


class DniTime:
    """A D'ni representation of a timestamp."""

    __slots__ = "_gahrtahvo", "_tahvo", "_gorahn", "_prorahn"

    if TYPE_CHECKING:
        _gahrtahvo: int
        _tahvo: int
        _gorahn: int
        _prorahn: int

        # Class properties
        min: "DniTime"
        max: "DniTime"
        resolution: DniTimedelta

    def __new__(
        cls,
        gahrtahvo: int = DNI_UNITS["gahrtahvo"]["min"],
        tahvo: int = DNI_UNITS["tahvo"]["min"],
        gorahn: int = DNI_UNITS["gorahn"]["min"],
        prorahn: int = DNI_UNITS["prorahn"]["min"],
    ) -> "DniTime":
        gahrtahvo, tahvo, gorahn, prorahn = check_dni_time_fields(
            gahrtahvo, tahvo, gorahn, prorahn
        )
        self = object.__new__(cls)
        self._gahrtahvo = gahrtahvo
        self._tahvo = tahvo
        self._gorahn = gorahn
        self._prorahn = prorahn
        return self

    @classmethod
    def from_isoformat(cls, time_str: str) -> "DniTime":
        if not isinstance(time_str, str):
            raise TypeError("from_isoformat: argument must be str.")

        try:
            assert 7 <= len(time_str) <= 11
            parts = time_str.split(":")
            if len(parts) != 4:
                raise ValueError(f"Invalid isoformat string: {time_str!r}")
            return cls(int(parts[0]), int(parts[1]), int(parts[2]), int(parts[3]))
        except Exception as exc:
            raise ValueError(f"Invalid isoformat string: {time_str!r}") from exc

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}"
            f"({self._gahrtahvo}, "
            f"{self._tahvo}, "
            f"{self._gorahn}, "
            f"{self._prorahn})"
        )

    def isoformat(self) -> str:
        """Return the date formatted according to ISO standard."""
        return (
            f"{self._gahrtahvo}:{self._tahvo:02}:"
            f"{self._gorahn:02}:{self._prorahn:02}"
        )

    __str__ = isoformat

    # Read-only field accessors
    @property
    def gahrtahvo(self) -> int:
        """gahrtahvo"""
        return self._gahrtahvo

    @property
    def tahvo(self) -> int:
        """tahvo"""
        return self._tahvo

    @property
    def gorahn(self) -> int:
        """gorahn"""
        return self._gorahn

    @property
    def prorahn(self) -> int:
        """prorahn"""
        return self._prorahn

    def replace(
        self,
        gahrtahvo: Optional[int] = None,
        tahvo: Optional[int] = None,
        gorahn: Optional[int] = None,
        prorahn: Optional[int] = None,
    ) -> "DniTime":
        """Return a new D'ni time with new values for the specified fields."""
        if gahrtahvo is None:
            gahrtahvo = self._gahrtahvo
        if tahvo is None:
            tahvo = self._tahvo
        if gorahn is None:
            gorahn = self._gorahn
        if prorahn is None:
            prorahn = self._prorahn
        return type(self)(gahrtahvo, tahvo, gorahn, prorahn)

    # Rich comparisons
    def _cmp(self, other: Any) -> int:
        """Helper function for rich comparison of dnidate objects."""
        # pylint: disable=protected-access
        assert isinstance(other, DniTime)
        mygopro = self._gorahn * 25 + self._prorahn
        othergopro = self._gorahn * 25 + self._prorahn
        return cmp(
            (self._gahrtahvo, self._tahvo, mygopro),
            (other._gahrtahvo, other._tahvo, othergopro),
        )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, DniTime):
            return self._cmp(other) == 0
        return NotImplemented

    def __le__(self, other: Any) -> bool:
        if isinstance(other, DniTime):
            return self._cmp(other) <= 0
        return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, DniTime):
            return self._cmp(other) < 0
        return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, DniTime):
            return self._cmp(other) >= 0
        return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, DniTime):
            return self._cmp(other) > 0
        return NotImplemented


DniTime.min = DniTime(0, 0, 0, 0)
DniTime.max = DniTime(4, 24, 24, 24)
DniTime.resolution = DniTimedelta(prorahntee=1)
